#ifndef PUGS_TRAITS_HPP
#define PUGS_TRAITS_HPP

#include <cstddef>
#include <memory>
#include <type_traits>
#include <variant>
#include <vector>

template <size_t N, typename T>
class TinyVector;
template <size_t N, typename T>
class TinyMatrix;

// Traits is_trivially_castable

template <typename T>
inline constexpr bool is_trivially_castable = std::is_trivial_v<T>;

template <size_t N, typename T>
inline constexpr bool is_trivially_castable<TinyVector<N, T>> = is_trivially_castable<T>;
template <size_t N, typename T>
inline constexpr bool is_trivially_castable<const TinyVector<N, T>> = is_trivially_castable<T>;

template <size_t N, typename T>
inline constexpr bool is_trivially_castable<TinyMatrix<N, T>> = is_trivially_castable<T>;
template <size_t N, typename T>
inline constexpr bool is_trivially_castable<const TinyMatrix<N, T>> = is_trivially_castable<T>;

// Traits is_false

template <typename T>
inline constexpr bool is_false_v = false;

// Traits is_shared_ptr

template <typename T>
inline constexpr bool is_shared_ptr_v = false;

template <typename T>
inline constexpr bool is_shared_ptr_v<std::shared_ptr<T>> = true;

// Traits is_shared_ptr

template <typename T>
inline constexpr bool is_unique_ptr_v = false;

template <typename T>
inline constexpr bool is_unique_ptr_v<std::unique_ptr<T>> = true;

// Traits is_weak_ptr

template <typename T>
inline constexpr bool is_weak_ptr_v = false;

template <typename T>
inline constexpr bool is_weak_ptr_v<std::weak_ptr<T>> = true;

// Traits is_std_ptr

template <typename T>
inline constexpr bool is_std_ptr_v = false;

template <typename T>
inline constexpr bool is_std_ptr_v<std::shared_ptr<T>> = true;
template <typename T>
inline constexpr bool is_std_ptr_v<std::unique_ptr<T>> = true;
template <typename T>
inline constexpr bool is_std_ptr_v<std::weak_ptr<T>> = true;

// Traits is_std_vector

template <typename T>
inline constexpr bool is_std_vector_v = false;

template <typename T>
inline constexpr bool is_std_vector_v<std::vector<T>> = true;

// Traits is_tiny_vector

template <typename T>
inline constexpr bool is_tiny_vector_v = false;

template <size_t N, typename T>
inline constexpr bool is_tiny_vector_v<TinyVector<N, T>> = true;

// Traits is_tiny_matrix

template <typename T>
inline constexpr bool is_tiny_matrix_v = false;

template <size_t N, typename T>
inline constexpr bool is_tiny_matrix_v<TinyMatrix<N, T>> = true;

// helper to check if a type is part of a variant

template <typename T, typename V>
struct is_variant;

template <typename T, typename... ALL_T>
struct is_variant<T, std::variant<ALL_T...>> : public std::disjunction<std::is_same<T, ALL_T>...>
{
};

#endif   // PUGS_TRAITS_HPP
